from typing import Tuple
import numpy as np
import scipy.ndimage
from pydicom.dataset import Dataset, FileMetaDataset
from pydicom.pixel_data_handlers import gdcm_handler, pillow_handler
from pydicom.uid import ImplicitVRLittleEndian, generate_uid
from pydicom._storage_sopclass_uids import SecondaryCaptureImageStorage


class Process():
    def __init__(self, data) -> None:
        self.__data = data
        self.__rgb = None
        self.__w1_min = self.__w1_max = None
        self.__w2_min = self.__w2_max = None
        self.__w3_min = self.__w3_max = None

    @property
    def rgb_windows(self) -> Tuple:
        return (
            self.__w1_min, self.__w1_max,
            self.__w2_min, self.__w2_max,
            self.__w3_min, self.__w3_max
        )

    @rgb_windows.setter
    def rgb_windows(self, *args: Tuple):
        self.__w1_min = args[0][0]
        self.__w1_max = args[0][1]
        self.__w2_min = args[0][2]
        self.__w2_max = args[0][3]
        self.__w3_min = args[0][4]
        self.__w3_max = args[0][5]

    @property
    def rgb(self):
        return self.__rgb

    def rgb_conversion(self):
        data_B = np.where(self.__data <= self.__w1_min,
                          self.__w1_min, self.__data)
        data_B = np.where(data_B >= self.__w1_max, self.__w1_min, data_B)
        data_B = (abs(data_B - self.__w1_min) /
                  abs(self.__w1_max - self.__w1_min)) * 255

        data_G = np.where(self.__data <= self.__w2_min,
                          self.__w2_min, self.__data)
        data_G = np.where(data_G >= self.__w2_max, self.__w2_min, data_G)
        data_G = (abs(data_G - self.__w2_min) /
                  abs(self.__w2_max - self.__w2_min)) * 255

        data_R = np.where(self.__data <= self.__w3_min,
                          self.__w3_min, self.__data)
        data_R = np.where(data_R >= self.__w3_max, self.__w3_min, data_R)
        data_R = (abs(data_R - self.__w3_min) /
                  abs(self.__w3_max - self.__w3_min)) * 255

        self.__rgb = np.zeros((*self.__data.shape, 3), dtype=np.uint8)
        self.__rgb[..., 0] = data_R.astype(np.uint8)
        self.__rgb[..., 1] = data_G.astype(np.uint8)
        self.__rgb[..., 2] = data_B.astype(np.uint8)

    def windowed_data(self, *args: int):
        result = np.where(self.__data <= args[0], args[0], self.__data)
        result = np.where(result >= args[1], args[1], result)
        return result

    def ct_get_pixels_hu(self, slices: np.ndarray) -> np.ndarray:
        """Get pixel data from dicom-file Dataset"""
        try:
            image = np.stack([s.pixel_array for s in slices])
        except RuntimeError:
            for slice_item in slices:
                slice_item.pixel_data_handlers = [gdcm_handler, pillow_handler]
            image = np.stack([s.pixel_array for s in slices])
        image = image.astype(np.int16)
        image[image == -2000] = 0
        for slice_number, _ in enumerate(slices):
            intercept = slices[slice_number].RescaleIntercept
            slope = slices[slice_number].RescaleSlope
            if slope != 1:
                image[slice_number] = slope * \
                    image[slice_number].astype(np.float64)
                image[slice_number] = image[slice_number].astype(np.int16)
            image[slice_number] += np.int16(intercept)
        return np.array(image, dtype=np.int16)

    def erosion(self, binary_image: np.ndarray) -> np.ndarray:
        """Morphological erosion"""
        return scipy.ndimage.binary_erosion(binary_image).astype(binary_image.dtype)

    def dilation(self, binary_image: np.ndarray) -> np.ndarray:
        """Morphological dilation"""
        return scipy.ndimage.binary_dilation(binary_image).astype(binary_image.dtype)

    def generate_sc_dataset(self, image, source_dataset, SeriesInstanceUID, number):
        meta = FileMetaDataset()
        meta.TransferSyntaxUID = ImplicitVRLittleEndian

        ds = Dataset()
        ds.file_meta = meta

        if hasattr(source_dataset, "PatientName"):
            ds.PatientName = source_dataset.PatientName
        else:
            ds.PatientName = "Unknown"

        if hasattr(source_dataset, "PatientID"):
            ds.PatientID = source_dataset.PatientID
        else:
            ds.PatientID = "Unknown"

        if hasattr(source_dataset, "PatientBirthDate"):
            ds.PatientBirthDate = source_dataset.PatientBirthDate
        else:
            ds.PatientBirthDate = "Unknown"

        if hasattr(source_dataset, "PatientSex"):
            ds.PatientSex = source_dataset.PatientSex
        else:
            ds.PatientSex = "O"

        if hasattr(source_dataset, "StudyDescription"):
            ds.StudyDescription = source_dataset.StudyDescription
        else:
            ds.StudyDescription = "default"

        ds.SeriesDescription = "SC Sciberia"
        ds.StudyInstanceUID = source_dataset.StudyInstanceUID

        if hasattr(source_dataset, "StudyDate"):
            ds.StudyDate = source_dataset.StudyDate
        else:
            ds.StudyDate = "19700101"

        if hasattr(source_dataset, "StudyTime"):
            ds.StudyTime = source_dataset.StudyTime
        else:
            ds.StudyTime = "000000"

        if hasattr(source_dataset, "ReferringPhysicianName"):
            ds.ReferringPhysicianName = source_dataset.ReferringPhysicianName
        else:
            ds.ReferringPhysicianName = None

        if hasattr(source_dataset, "StudyID"):
            ds.StudyID = source_dataset.StudyID
        else:
            ds.StudyID = None

        if hasattr(source_dataset, "AccessionNumber"):
            ds.AccessionNumber = source_dataset.AccessionNumber
        else:
            ds.AccessionNumber = None

        ds.Modality = "SC"
        ds.SeriesInstanceUID = SeriesInstanceUID
        ds.SeriesNumber = 100
        ds.ConversionType = "DF"

        ds.InstanceNumber = number

        arr = np.stack([image])
        ds.NumberOfFrames, ds.Rows, ds.Columns, ds.SamplesPerPixel = arr.shape
        ds.PhotometricInterpretation = "RGB"
        ds.BitsAllocated = 8
        ds.BitsStored = 8
        ds.HighBit = 7
        ds.PixelRepresentation = 0
        ds.PlanarConfiguration = 0
        ds.PixelData = arr.tobytes()

        ds.SOPClassUID = SecondaryCaptureImageStorage
        ds.SOPInstanceUID = generate_uid()

        return ds



import errno
from itertools import groupby
import os
from pathlib import Path
import pydicom
from typing import Dict, List, Tuple


def commonprefix(args):
    return os.path.commonpath(args)


def series_projection(val):
    return val.SeriesInstanceUID


def study_projection(val):
    return val.StudyInstanceUID


class Reader():
    def __init__(self, path) -> None:
        self.__path = path
        self.__filenames = None
    
    @property
    def filenames(self) -> List:
        return self.__filenames

    def read_filenames(self) -> None:
        studies_candidates = []
        for root, _, files in os.walk(self.__path):
            scan = None
            if "DICOMDIR" in files:
                scan = {
                    "path": root,
                    "dicomdir": True
                }
            else:
                dicoms_list_candidates = self.dicoms_list_in_dir(root)
                if len(studies_candidates) > 0:
                    _sc = [item["path"] for item in studies_candidates]
                    if _sc[-1] in root:
                        continue
                if len(dicoms_list_candidates) > 0:
                    scan = {
                        "path": root,
                        "dicomdir": False
                    }
            if not scan is None:
                studies_candidates.append(scan)
        datasets = [
            {
                "data": pydicom.dcmread(os.path.join(study_candidate["path"], file), specific_tags=["StudyInstanceUID", "SeriesInstanceUID"]),
                "path": study_candidate["path"]
            }
            for study_candidate in studies_candidates if not study_candidate["dicomdir"]
            for file in os.listdir(study_candidate["path"]) if self.is_dicom(os.path.join(study_candidate["path"], file))
        ]
        groupped_studies = [
            [list(i)
             for _, i in groupby(k, lambda m: m["data"].SeriesInstanceUID)]
            for k in [list(j) for _, j in groupby(datasets, lambda m: m["data"].StudyInstanceUID)]
        ]
        dicomdir_studies_candidates = [
            i for i in studies_candidates if i["dicomdir"]]
        dcm_studies_candidates = []
        for studies in groupped_studies:
            paths = []
            for series in studies:
                for item in series:
                    paths.append(item["path"])
            dcm_studies_candidates.append({
                "path": commonprefix(paths),
                "dicomdir": False
            })
        self.__filenames = dicomdir_studies_candidates + dcm_studies_candidates

    def read_studies_generator(self, stop_before_pixels: bool = False) -> List:
        for study_struct in self.__filenames:
            datasets = []
            path = study_struct["path"]
            if study_struct["dicomdir"]:
                dicomdir = pydicom.dcmread(os.path.join(path, "DICOMDIR"))
                for patient_record in dicomdir.patient_records:
                    studies = [
                        ii for ii in patient_record.children if ii.DirectoryRecordType == "STUDY"
                    ]
                    for study in studies:
                        all_series = [
                            ii for ii in study.children if ii.DirectoryRecordType == "SERIES"
                        ]
                        for series in all_series:
                            images = [
                                ii for ii in series.children
                                if ii.DirectoryRecordType == "IMAGE"
                            ]
                            elems = [ii["ReferencedFileID"] for ii in images]
                            paths = [[ee.value] if ee.VM ==
                                    1 else ee.value for ee in elems]
                            paths = [Path(*p) for p in paths]
                            _datasets = [
                                pydicom.dcmread(os.path.join(path, image_path),
                                                stop_before_pixels=stop_before_pixels)
                                for image_path in paths
                            ]
                            datasets += _datasets
            else:
                p = Path(path)
                dicoms = [x for x in p.glob("**/*") if self.is_dicom(x)]
                datasets = [
                    pydicom.dcmread(
                        str(dicom), stop_before_pixels=stop_before_pixels)
                    for dicom in dicoms
                ]
            for data in datasets:
                if not hasattr(data, "StudyDescription"):
                    data.StudyDescription = "default"
                if not hasattr(data, "SeriesDescription"):
                    data.SeriesDescription = "default"
            groupped_studies = []
            for item in [
                list(it) for k, it in groupby(datasets, study_projection)
            ]:
                groupped_by_series = [
                    list(it) for k, it in groupby(item, series_projection)
                ]
                groupped_studies.append(groupped_by_series)
            for series in groupped_studies[0]:
                if len(series) > 1:
                    if hasattr(series[0], "ImagePositionPatient"):
                        series.sort(key=lambda x: float(
                            x.ImagePositionPatient[2]), reverse=True)
            yield groupped_studies[0]

    def checkpath(self, path: str) -> None:
        """Check path for existing"""
        if not os.path.exists(path):
            try:
                os.makedirs(path, exist_ok=True)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise

    def get_uids(self, path: str) -> Tuple[str, str]:
        """Get StudyInstanceUID and SeriesInstanceUID, no pixel data"""
        study_data = pydicom.read_file(path, stop_before_pixels=True)
        return study_data.StudyInstanceUID, study_data.SeriesInstanceUID

    def is_dicom(self, path: str) -> bool:
        """Check file whether dicom-file or not"""
        if not os.path.isfile(path):
            return False
        try:
            with open(path, "rb") as file_name:
                return file_name.read(132).decode("ASCII")[-4:] == "DICM"
        except UnicodeDecodeError:
            return False

    def dicoms_list_in_dir(self, path: str = ".") -> List[str]:
        """Forms list of dicom-files in directory"""
        path = os.path.expanduser(path)
        candidates = [os.path.join(path, f) for f in sorted(os.listdir(path))]
        return [f for f in candidates if self.is_dicom(f)]

    def is_dicomdir(self, path: str = ".") -> Tuple[bool, str]:
        """Find first DICOMDIR in subdirectories"""
        dicomdir = False
        for root, _, files in os.walk(path):
            if "DICOMDIR" in files:
                return True, root
        if not dicomdir:
            return False, path

    def batch_reader(self, scanpath: str) -> List[Dict]:
        """Recursively read files and subdirectories to find dicom-file collections"""
        scans = []
        for root, _, files in os.walk(scanpath):
            dicoms_list_candidates = self.dicoms_list_in_dir(root)
            scan = {}
            if len(dicoms_list_candidates) > 0:
                scan["nrrd"] = [file for file in files if ".nrrd" in file]
                scan["path"] = root
                scans.append(scan)
        return scans

    def dicomdir_reader(self, path: str) -> List:
        """Read dicomdir"""
        dicomdir = self.read_dicomdir(os.path.join(path, "DICOMDIR"))
        datasets = []
        for patient_record in dicomdir.patient_records:
            studies = patient_record.children
            for study in studies:
                all_series = study.children
                for series in all_series:
                    if "SeriesDescription" not in series:
                        series.SeriesDescription = "default"
                    image_records = [child for child in series.children]
                    image_filenames = [os.path.join(
                        path, *image_rec.ReferencedFileID) for image_rec in image_records]
                    dataset = [pydicom.dcmread(image_filename)
                               for image_filename in image_filenames]
                    datasets.append(dataset)
        datasets = datasets[0]
        return datasets



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="sciberia",
    version="0.1.1",
    description="Sciberia helper libraries",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/sciberia-llc/sciberia",
    author="Sciberia, LLC",
    author_email="info@sciberia.io",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["sciberia"],
    include_package_data=True,
    install_requires=[
        "numpy",
        "pydicom",
        "pytest",
        "scipy"
    ]
)
